// Copyright (c) Microsoft Open Technologies, Inc. All rights reserved.
// Licensed under the Apache License, Version 2.0. See License.txt in the project root for license information.

using System;
using System.Collections.Generic;
using System.Linq;
using System.Threading;
using System.Threading.Tasks;
using Microsoft.Data.Entity.FunctionalTests.TestModels.Northwind;
using Microsoft.Data.Entity.Query;
using Xunit;

namespace Microsoft.Data.Entity.FunctionalTests
{
    public abstract class AsyncQueryTestBase<TFixture> : IClassFixture<TFixture>
        where TFixture : NorthwindQueryFixtureBase, new()
    {
        [Fact]
        public virtual async Task Queryable_simple()
        {
            Assert.Equal(91,
                await AssertQuery<Customer>(cs => cs));
        }

        [Fact]
        public virtual async Task Queryable_simple_anonymous()
        {
            Assert.Equal(91,
                await AssertQuery<Customer>(cs => cs.Select(c => new { c })));
        }

        [Fact]
        public virtual async Task Queryable_nested_simple()
        {
            Assert.Equal(91,
                await AssertQuery<Customer>(cs =>
                    from c1 in (from c2 in (from c3 in cs select c3) select c2) select c1));
        }

        [Fact]
        public virtual async Task Take_simple()
        {
            Assert.Equal(10,
                await AssertQuery<Customer>(cs => cs.OrderBy(c => c.CustomerID).Take(10), assertOrder: true));
        }

        [Fact]
        public virtual async Task Take_simple_projection()
        {
            Assert.Equal(10,
                await AssertQuery<Customer>(cs => cs.OrderBy(c => c.CustomerID).Select(c => c.City).Take(10), assertOrder: true));
        }

        [Fact]
        public virtual async Task Skip()
        {
            await AssertQuery<Customer>(cs => cs.OrderBy(c => c.CustomerID).Skip(5), assertOrder: true);
        }

        [Fact]
        public virtual async Task Take_Skip()
        {
            await AssertQuery<Customer>(cs => cs.OrderBy(c => c.ContactName).Take(10).Skip(5), assertOrder: true);
        }

        [Fact]
        public virtual async Task Distinct_Skip()
        {
            await AssertQuery<Customer>(cs =>
                cs.Distinct().OrderBy(c => c.ContactName).Skip(5), assertOrder: true);
        }

        [Fact]
        public virtual async Task Skip_Take()
        {
            await AssertQuery<Customer>(cs =>
                cs.OrderBy(c => c.ContactName).Skip(5).Take(10), assertOrder: true);
        }

        [Fact]
        public virtual async Task Distinct_Skip_Take()
        {
            await AssertQuery<Customer>(cs =>
                cs.Distinct().OrderBy(c => c.ContactName).Skip(5).Take(10), assertOrder: true);
        }

        [Fact]
        public virtual async Task Skip_Distinct()
        {
            await AssertQuery<Customer>(cs =>
                cs.OrderBy(c => c.ContactName).Skip(5).Distinct());
        }

        [Fact]
        public virtual async Task Skip_Take_Distinct()
        {
            await AssertQuery<Customer>(cs =>
                cs.OrderBy(c => c.ContactName).Skip(5).Take(10).Distinct());
        }

        [Fact]
        public virtual async Task Take_Skip_Distinct()
        {
            await AssertQuery<Customer>(cs =>
                cs.OrderBy(c => c.ContactName).Take(10).Skip(5).Distinct());
        }

        [Fact]
        public virtual async Task Take_Distinct()
        {
            await AssertQuery<Order>(os => os.OrderBy(o => o.OrderID).Take(5).Distinct());
        }

        [Fact]
        public virtual async Task Distinct_Take()
        {
            await AssertQuery<Order>(os => os.Distinct().OrderBy(o => o.OrderID).Take(5), assertOrder: true);
        }

        [Fact]
        public virtual async Task Distinct_Take_Count()
        {
            await AssertQuery<Order>(os => os.Distinct().Take(5).CountAsync());
        }

        [Fact]
        public virtual async Task Take_Distinct_Count()
        {
            await AssertQuery<Order>(os => os.Take(5).Distinct().CountAsync());
        }

        [Fact]
        public virtual async Task Any_simple()
        {
            await AssertQuery<Customer>(cs => cs.AnyAsync());
        }

        [Fact]
        public virtual async Task Any_predicate()
        {
            await AssertQuery<Customer>(cs => cs.AnyAsync(c => c.ContactName.StartsWith("A")));
        }

        [Fact]
        public virtual async Task All_top_level()
        {
            await AssertQuery<Customer>(cs => cs.AllAsync(c => c.ContactName.StartsWith("A")));
        }

        [Fact]
        public virtual async Task All_top_level_subquery()
        {
            // ReSharper disable once PossibleUnintendedReferenceComparison
            await AssertQuery<Customer>(cs => cs.AllAsync(c1 => cs.Any(c2 => cs.Any(c3 => c1 == c3))));
        }

        [Fact]
        public virtual async Task Select_into()
        {
            await AssertQuery<Customer>(cs =>
                from c in cs
                select c.CustomerID
                into id
                where id == "ALFKI"
                select id);
        }

        [Fact]
        public virtual async Task Projection_when_null_value()
        {
            await AssertQuery<Customer>(cs => cs.Select(c => c.Region));
        }

        [Fact]
        public virtual async Task Take_with_single()
        {
            await AssertQuery<Customer>(cs => cs.OrderBy(c => c.CustomerID).Take(1).SingleAsync());
        }

        [Fact]
        public virtual async Task Take_with_single_select_many()
        {
            await AssertQuery<Customer, Order>((cs, os) =>
                (from c in cs
                    from o in os
                    orderby c.CustomerID, o.OrderID
                    select new { c, o })
                    .Take(1)
                    .Cast<object>()
                    .SingleAsync());
        }

        [Fact]
        public virtual async Task Where_simple()
        {
            await AssertQuery<Customer>(cs => cs.Where(c => c.City == "London"));
        }

        [Fact]
        public virtual async Task Where_simple_shadow()
        {
            await AssertQuery<Employee>(es => es.Where(e => e.Property<string>("Title") == "Sales Representative"));
        }

        [Fact]
        public virtual async Task Where_simple_shadow_projection()
        {
            await AssertQuery<Employee>(
                es => es.Where(e => e.Property<string>("Title") == "Sales Representative")
                    .Select(e => e.Property<string>("Title")));
        }

        [Fact]
        public virtual async Task Where_client()
        {
            await AssertQuery<Customer>(cs => cs.Where(c => c.IsLondon));
        }

        [Fact]
        public virtual async Task First_client_predicate()
        {
            await AssertQuery<Customer>(cs => cs.OrderBy(c => c.CustomerID).FirstAsync(c => c.IsLondon));
        }

        [Fact]
        public virtual async Task Where_equals_method_string()
        {
            await AssertQuery<Customer>(cs => cs.Where(c => c.City.Equals("London")));
        }

        [Fact]
        public virtual async Task Where_equals_method_int()
        {
            await AssertQuery<Employee>(es => es.Where(e => e.EmployeeID.Equals(1)));
        }

        [Fact]
        public virtual async Task Where_comparison_nullable_type_not_null()
        {
            await AssertQuery<Employee>(es => es.Where(e => e.ReportsTo == 2));
        }

        [Fact]
        public virtual async Task Where_comparison_nullable_type_null()
        {
            await AssertQuery<Employee>(es => es.Where(e => e.ReportsTo == null));
        }

        [Fact]
        public virtual async Task Where_string_length()
        {
            await AssertQuery<Customer>(cs => cs.Where(c => c.City.Length == 6));
        }

        [Fact]
        public virtual async Task Where_simple_reversed()
        {
            await AssertQuery<Customer>(cs => cs.Where(c => "London" == c.City));
        }

        [Fact]
        public virtual async Task Where_is_null()
        {
            await AssertQuery<Customer>(cs => cs.Where(c => c.City == null));
        }

        [Fact]
        public virtual async Task Where_null_is_null()
        {
            // ReSharper disable once EqualExpressionComparison
            await AssertQuery<Customer>(cs => cs.Where(c => null == null));
        }

        [Fact]
        public virtual async Task Where_constant_is_null()
        {
            await AssertQuery<Customer>(cs => cs.Where(c => "foo" == null));
        }

        [Fact]
        public virtual async Task Where_is_not_null()
        {
            await AssertQuery<Customer>(cs => cs.Where(c => c.City != null));
        }

        [Fact]
        public virtual async Task Where_null_is_not_null()
        {
            // ReSharper disable once EqualExpressionComparison
            await AssertQuery<Customer>(cs => cs.Where(c => null != null));
        }

        [Fact]
        public virtual async Task Where_constant_is_not_null()
        {
            await AssertQuery<Customer>(cs => cs.Where(c => "foo" != null));
        }

        [Fact]
        public virtual async Task Where_identity_comparison()
        {
            // ReSharper disable once EqualExpressionComparison
            await AssertQuery<Customer>(cs => cs.Where(c => c.City == c.City));
        }

        [Fact]
        public virtual async Task Where_select_many_or()
        {
            await AssertQuery<Customer, Employee>((cs, es) =>
                from c in cs
                from e in es
                where c.City == "London"
                      || e.City == "London"
                select new { c, e });
        }

        [Fact]
        public virtual async Task Where_select_many_or2()
        {
            await AssertQuery<Customer, Employee>((cs, es) =>
                from c in cs
                from e in es
                where c.City == "London"
                      || c.City == "Berlin"
                select new { c, e });
        }

        [Fact]
        public virtual async Task Where_select_many_and()
        {
            await AssertQuery<Customer, Employee>((cs, es) =>
                from c in cs
                from e in es
                where (c.City == "London" && c.Country == "UK")
                      && (e.City == "London" && e.Country == "UK")
                select new { c, e });
        }

        [Fact]
        public virtual async Task Where_primitive()
        {
            await AssertQuery<Employee>(es =>
                es.Select(e => e.EmployeeID).Take(9).Where(i => i == 5));
        }

        [Fact]
        public virtual async Task Where_true()
        {
            Assert.Equal(91,
                await AssertQuery<Customer>(cs => cs.Where(c => true)));
        }

        [Fact]
        public virtual async Task Where_false()
        {
            Assert.Equal(0,
                await AssertQuery<Customer>(cs => cs.Where(c => false)));
        }

        //        TODO: Re-write entity ref equality to identity equality.
        //
        //        [Fact]
        //        public virtual async Task Where_compare_entity_equal()
        //        {
        //            var alfki = NorthwindData.Customers.Single(c => c.CustomerID == "ALFKI");
        //
        //            Assert.Equal(1,
        //                // ReSharper disable once PossibleUnintendedReferenceComparison
        //                await AssertQuery<Customer>(cs => cs.Where(c => c == alfki)));
        //        }
        //
        //        [Fact]
        //        public virtual async Task Where_compare_entity_not_equal()
        //        {
        //            var alfki = new Customer { CustomerID = "ALFKI" };
        //
        //            Assert.Equal(90,
        //                // ReSharper disable once PossibleUnintendedReferenceComparison
        //                await AssertQuery<Customer>(cs => cs.Where(c => c != alfki)));
        //
        //        [Fact]
        //        public virtual async Task Project_compare_entity_equal()
        //        {
        //            var alfki = NorthwindData.Customers.Single(c => c.CustomerID == "ALFKI");
        //
        //            Assert.Equal(1,
        //                // ReSharper disable once PossibleUnintendedReferenceComparison
        //                await AssertQuery<Customer>(cs => cs.Select(c => c == alfki)));
        //        }
        //
        //        [Fact]
        //        public virtual async Task Project_compare_entity_not_equal()
        //        {
        //            var alfki = new Customer { CustomerID = "ALFKI" };
        //
        //            Assert.Equal(90,
        //                // ReSharper disable once PossibleUnintendedReferenceComparison
        //                await AssertQuery<Customer>(cs => cs.Select(c => c != alfki)));
        //        }

        [Fact]
        public virtual async Task Where_compare_constructed_equal()
        {
            await AssertQuery<Customer>(cs =>
                cs.Where(c => new { x = c.City } == new { x = "London" }));
        }

        [Fact]
        public virtual async Task Where_compare_constructed_multi_value_equal()
        {
            await AssertQuery<Customer>(cs =>
                cs.Where(c => new { x = c.City, y = c.Country } == new { x = "London", y = "UK" }));
        }

        [Fact]
        public virtual async Task Where_compare_constructed_multi_value_not_equal()
        {
            await AssertQuery<Customer>(cs =>
                cs.Where(c => new { x = c.City, y = c.Country } != new { x = "London", y = "UK" }));
        }

        [Fact]
        public virtual async Task Where_compare_constructed()
        {
            await AssertQuery<Customer>(cs =>
                cs.Where(c => new { x = c.City } == new { x = "London" }));
        }

        [Fact]
        public virtual async Task Where_projection()
        {
            await AssertQuery<Customer>(cs => cs.Where(c => c.City == "London").Select(c => c.CompanyName));
        }

        [Fact]
        public virtual async Task Select_scalar()
        {
            await AssertQuery<Customer>(cs => cs.Select(c => c.City));
        }

        [Fact]
        public virtual async Task Select_anonymous_one()
        {
            await AssertQuery<Customer>(cs => cs.Select(c => new { c.City }));
        }

        [Fact]
        public virtual async Task Select_anonymous_two()
        {
            await AssertQuery<Customer>(cs => cs.Select(c => new { c.City, c.Phone }));
        }

        [Fact]
        public virtual async Task Select_anonymous_three()
        {
            await AssertQuery<Customer>(cs => cs.Select(c => new { c.City, c.Phone, c.Country }));
        }

        [Fact]
        public virtual async Task Select_customer_table()
        {
            await AssertQuery<Customer>(cs => cs);
        }

        [Fact]
        public virtual async Task Select_customer_identity()
        {
            await AssertQuery<Customer>(cs => cs.Select(c => c));
        }

        [Fact]
        public virtual async Task Select_anonymous_with_object()
        {
            await AssertQuery<Customer>(cs => cs.Select(c => new { c.City, c }));
        }

        [Fact]
        public virtual async Task Select_anonymous_nested()
        {
            await AssertQuery<Customer>(cs => cs.Select(c => new { c.City, Country = new { c.Country } }));
        }

        [Fact]
        public virtual async Task Select_anonymous_empty()
        {
            await AssertQuery<Customer>(cs => cs.Select(c => new { }));
        }

        [Fact]
        public virtual async Task Select_anonymous_literal()
        {
            await AssertQuery<Customer>(cs => cs.Select(c => new { X = 10 }));
        }

        [Fact]
        public virtual async Task Select_constant_int()
        {
            await AssertQuery<Customer>(cs => cs.Select(c => 0));
        }

        [Fact]
        public virtual async Task Select_constant_null_string()
        {
            await AssertQuery<Customer>(cs => cs.Select(c => (string)null));
        }

        [Fact]
        public virtual async Task Select_local()
        {
            // ReSharper disable once ConvertToConstant.Local
            var x = 10;

            await AssertQuery<Customer>(cs => cs.Select(c => x));
        }

        [Fact]
        public virtual async Task Select_scalar_primitive()
        {
            Assert.Equal(9,
                await AssertQuery<Employee>(es => es.Select(e => e.EmployeeID)));
        }

        [Fact]
        public virtual async Task Select_scalar_primitive_after_take()
        {
            Assert.Equal(9,
                await AssertQuery<Employee>(es => es.Take(9).Select(e => e.EmployeeID)));
        }

        [Fact]
        public virtual async Task Select_project_filter()
        {
            await AssertQuery<Customer>(cs =>
                from c in cs
                where c.City == "London"
                select c.CompanyName);
        }

        [Fact]
        public virtual async Task Select_project_filter2()
        {
            await AssertQuery<Customer>(cs =>
                from c in cs
                where c.City == "London"
                select c.City);
        }

        [Fact]
        public virtual async Task Select_nested_collection()
        {
            await AssertQuery<Customer, Order>((cs, os) =>
                from c in cs
                where c.City == "London"
                orderby c.CustomerID
                select os
                    .Where(o => o.CustomerID == c.CustomerID
                                && o.OrderDate.Value.Year == 1997)
                    .Select(o => o.OrderID)
                    .OrderBy(o => o),
                asserter:
                    (l2oResults, efResults) =>
                        {
                            var l2oObjects
                                = l2oResults
                                    .SelectMany(q1 => ((IEnumerable<int>)q1));

                            var efObjects
                                = efResults
                                    .SelectMany(q1 => ((IEnumerable<int>)q1));

                            Assert.Equal(l2oObjects, efObjects);
                        });
        }

        [Fact]
        public virtual async Task Select_correlated_subquery_projection()
        {
            await AssertQuery<Customer, Order>((cs, os) =>
                from c in cs
                select os
                    .Where(o => o.CustomerID == c.CustomerID),
                asserter:
                    (l2oResults, efResults) =>
                        {
                            var l2oObjects
                                = l2oResults
                                    .SelectMany(q1 => ((IEnumerable<Order>)q1))
                                    .OrderBy(o => o.OrderID);

                            var efObjects
                                = efResults
                                    .SelectMany(q1 => ((IEnumerable<Order>)q1))
                                    .OrderBy(o => o.OrderID);

                            Assert.Equal(l2oObjects, efObjects);
                        });
        }

        [Fact]
        public virtual async Task Select_correlated_subquery_ordered()
        {
            await AssertQuery<Customer, Order>((cs, os) =>
                from c in cs
                select os
                    .OrderBy(o => c.CustomerID),
                asserter:
                    (l2oResults, efResults) =>
                        {
                            var l2oObjects
                                = l2oResults
                                    .SelectMany(q1 => ((IEnumerable<Order>)q1))
                                    .OrderBy(o => o.OrderID);

                            var efObjects
                                = efResults
                                    .SelectMany(q1 => ((IEnumerable<Order>)q1))
                                    .OrderBy(o => o.OrderID);

                            Assert.Equal(l2oObjects, efObjects);
                        });
        }

        // TODO: Re-linq parser
        // [Fact]
        // public virtual async Task Select_nested_ordered_enumerable_collection()
        // {
        //     await AssertQuery<Customer>(cs =>
        //         cs.Select(c => cs.AsEnumerable().OrderBy(c2 => c2.CustomerID)),
        //         assertOrder: true);
        // }

        [Fact]
        public virtual async Task Select_nested_collection_in_anonymous_type()
        {
            await AssertQuery<Customer, Order>((cs, os) =>
                from c in cs
                where c.CustomerID == "ALFKI"
                select new
                    {
                        CustomerId = c.CustomerID,
                        OrderIds
                            = os.Where(o => o.CustomerID == c.CustomerID
                                            && o.OrderDate.Value.Year == 1997)
                                .Select(o => o.OrderID)
                                .OrderBy(o => o),
                        Customer = c
                    },
                asserter:
                    (l2oResults, efResults) =>
                        {
                            dynamic l2oResult = l2oResults.Single();
                            dynamic efResult = efResults.Single();

                            Assert.Equal(l2oResult.CustomerId, efResult.CustomerId);
                            Assert.Equal((IEnumerable<int>)l2oResult.OrderIds, (IEnumerable<int>)efResult.OrderIds);
                            Assert.Equal(l2oResult.Customer, efResult.Customer);
                        });
        }

        [Fact]
        public virtual async Task Select_subquery_recursive_trivial()
        {
            await AssertQuery<Employee>(es =>
                from e1 in es
                select (from e2 in es
                    select (from e3 in es
                        orderby e3.EmployeeID
                        select e3)),
                asserter:
                    (l2oResults, efResults) =>
                        {
                            var l2oObjects
                                = l2oResults
                                    .SelectMany(q1 => ((IEnumerable<object>)q1)
                                        .SelectMany(q2 => (IEnumerable<object>)q2));

                            var efObjects
                                = efResults
                                    .SelectMany(q1 => ((IEnumerable<object>)q1)
                                        .SelectMany(q2 => (IEnumerable<object>)q2));

                            Assert.Equal(l2oObjects, efObjects);
                        });
        }

        // TODO: [Fact] See #153
        public virtual async Task Where_subquery_on_collection()
        {
            await AssertQuery<Product, OrderDetail>((pr, od) =>
                from p in pr
                where p.OrderDetails.Contains(od.FirstOrDefault(orderDetail => orderDetail.Discount == 0.1))
                select p);
        }

        [Fact]
        public virtual async Task Where_query_composition()
        {
            await AssertQuery<Employee>(es =>
                from e1 in es
                where e1.FirstName == es.OrderBy(e => e.EmployeeID).First().FirstName
                select e1);
        }

        [Fact]
        public virtual async Task Where_subquery_recursive_trivial()
        {
            await AssertQuery<Employee>(es =>
                from e1 in es
                where (from e2 in es
                    where (from e3 in es
                        orderby e3.EmployeeID
                        select e3).Any()
                    select e2).Any()
                orderby e1.EmployeeID
                select e1,
                assertOrder: true);
        }

        [Fact]
        public virtual async Task Select_nested_collection_deep()
        {
            await AssertQuery<Customer, Order>((cs, os) =>
                from c in cs
                where c.City == "London"
                orderby c.CustomerID
                select (from o1 in os
                    where o1.CustomerID == c.CustomerID
                          && o1.OrderDate.Value.Year == 1997
                    orderby o1.OrderID
                    select (from o2 in os
                        where o1.CustomerID == c.CustomerID
                        orderby o2.OrderID
                        select o1.OrderID)),
                asserter:
                    (l2oResults, efResults) =>
                        {
                            var l2oObjects
                                = l2oResults
                                    .SelectMany(q1 => ((IEnumerable<object>)q1)
                                        .SelectMany(q2 => (IEnumerable<int>)q2));

                            var efObjects
                                = efResults
                                    .SelectMany(q1 => ((IEnumerable<object>)q1)
                                        .SelectMany(q2 => (IEnumerable<int>)q2));

                            Assert.Equal(l2oObjects, efObjects);
                        });
        }

        [Fact]
        public virtual async Task OrderBy_scalar_primitive()
        {
            await AssertQuery<Employee>(es =>
                es.Select(e => e.EmployeeID).OrderBy(i => i),
                assertOrder: true);
        }

        [Fact]
        public virtual async Task SelectMany_mixed()
        {
            await AssertQuery<Employee, Customer>(
                (es, cs) => from e1 in es
                    from s in new[] { "a", "b" }
                    from c in cs
                    select new { e1, s, c });
        }

        [Fact]
        public virtual async Task SelectMany_simple1()
        {
            await AssertQuery<Employee, Customer>(
                (es, cs) => from e in es
                    from c in cs
                    select new { c, e });
        }

        [Fact]
        public virtual async Task SelectMany_simple2()
        {
            await AssertQuery<Employee, Customer>(
                (es, cs) => from e1 in es
                    from c in cs
                    from e2 in es
                    select new { e1, c, e2.FirstName });
        }

        [Fact]
        public virtual async Task SelectMany_entity_deep()
        {
            await AssertQuery<Employee>(
                es =>
                    from e1 in es
                    from e2 in es
                    from e3 in es
                    select new { e2, e3, e1 });
        }

        [Fact]
        public virtual async Task SelectMany_projection1()
        {
            await AssertQuery<Employee>(
                es => from e1 in es
                    from e2 in es
                    select new { e1.City, e2.Country });
        }

        [Fact]
        public virtual async Task SelectMany_projection2()
        {
            await AssertQuery<Employee>(
                es => from e1 in es
                    from e2 in es
                    from e3 in es
                    select new { e1.City, e2.Country, e3.FirstName });
        }

        [Fact]
        public virtual async Task SelectMany_nested_simple()
        {
            await AssertQuery<Customer>(cs =>
                from c in cs
                from c1 in
                    (from c2 in (from c3 in cs select c3) select c2)
                orderby c1.CustomerID
                select c1,
                assertOrder: true);
        }

        [Fact]
        public virtual async Task SelectMany_correlated_simple()
        {
            await AssertQuery<Customer, Employee>((cs, es) =>
                from c in cs
                from e in es
                where c.City == e.City
                orderby c.CustomerID, e.EmployeeID
                select new { c, e },
                assertOrder: true);
        }

        [Fact]
        public virtual async Task SelectMany_correlated_subquery_simple()
        {
            await AssertQuery<Customer, Employee>(
                (cs, es) =>
                    from c in cs
                    from e in es.Where(e => e.City == c.City)
                    orderby c.CustomerID, e.EmployeeID
                    select new { c, e },
                assertOrder: true);
        }

        [Fact]
        public virtual async Task SelectMany_correlated_subquery_hard()
        {
            await AssertQuery<Customer, Employee>(
                (cs, es) =>
                    from c1 in
                        (from c2 in cs.Take(91) select c2.City).Distinct()
                    from e1 in
                        (from e2 in es where c1 == e2.City select new { e2.City, c1 }).Take(9)
                    from e2 in
                        (from e3 in es where e1.City == e3.City select c1).Take(9)
                    select new { c1, e1 });
        }

        [Fact]
        public virtual async Task SelectMany_cartesian_product_with_ordering()
        {
            await AssertQuery<Customer, Employee>(
                (cs, es) =>
                    from c in cs
                    from e in es
                    where c.City == e.City
                    orderby e.City ascending, c.CustomerID descending
                    select new { c, e.City },
                assertOrder: true);
        }

        [Fact]
        public virtual async Task SelectMany_primitive()
        {
            await AssertQuery<Employee>(
                es =>
                    from e1 in es
                    from i in es.Select(e2 => e2.EmployeeID)
                    select i);
        }

        [Fact]
        public virtual async Task SelectMany_primitive_select_subquery()
        {
            await AssertQuery<Employee>(
                es =>
                    from e1 in es
                    from i in es.Select(e2 => e2.EmployeeID)
                    select es.Any());
        }

        [Fact]
        public virtual async Task Join_customers_orders_projection()
        {
            await AssertQuery<Customer, Order>((cs, os) =>
                from c in cs
                join o in os on c.CustomerID equals o.CustomerID
                select new { c.ContactName, o.OrderID });
        }

        [Fact]
        public virtual async Task Join_customers_orders_entities()
        {
            await AssertQuery<Customer, Order>((cs, os) =>
                from c in cs
                join o in os on c.CustomerID equals o.CustomerID
                select new { c, o });
        }

        [Fact]
        public virtual async Task Join_select_many()
        {
            await AssertQuery<Customer, Order, Employee>((cs, os, es) =>
                from c in cs
                join o in os on c.CustomerID equals o.CustomerID
                from e in es
                select new { c, o, e });
        }

        [Fact]
        public virtual async Task Join_customers_orders_select()
        {
            await AssertQuery<Customer, Order>((cs, os) =>
                from c in cs
                join o in os on c.CustomerID equals o.CustomerID
                select new { c.ContactName, o.OrderID }
                into p
                select p);
        }

        [Fact]
        public virtual async Task Join_customers_orders_with_subquery()
        {
            await AssertQuery<Customer, Order>((cs, os) =>
                from c in cs
                join o1 in
                    (from o2 in os orderby o2.OrderID select o2) on c.CustomerID equals o1.CustomerID
                where o1.CustomerID == "ALFKI"
                select new { c.ContactName, o1.OrderID });
        }

        [Fact]
        public virtual async Task Join_composite_key()
        {
            await AssertQuery<Customer, Order>((cs, os) =>
                from c in cs
                join o in os on new { a = c.CustomerID, b = c.CustomerID }
                    equals new { a = o.CustomerID, b = o.CustomerID }
                select new { c, o });
        }

        [Fact]
        public virtual async Task Join_client_new_expression()
        {
            await AssertQuery<Customer, Order>((cs, os) =>
                from c in cs
                join o in os on new Foo { Bar = c.CustomerID } equals new Foo { Bar = o.CustomerID }
                select new { c, o });
        }

        private class Foo
        {
            public string Bar { get; set; }
        }

        //[Fact]
        // TODO: Convert orders to IEnumerable<T>
        public virtual async Task GroupJoin_customers_orders()
        {
            await AssertQuery<Customer, Order>((cs, os) =>
                from c in cs
                join o in os.OrderBy(o => o.OrderID) on c.CustomerID equals o.CustomerID into orders
                select new { customer = c, orders = orders.ToList() },
                asserter: (l2oItems, efItems) =>
                    {
                        foreach (var pair in
                            from dynamic l2oItem in l2oItems
                            join dynamic efItem in efItems on l2oItem.customer equals efItem.customer
                            select new { l2oItem, efItem })
                        {
                            Assert.Equal(pair.l2oItem.orders, pair.efItem.orders);
                        }
                    });
        }

        [Fact]
        public virtual async Task GroupJoin_customers_orders_count()
        {
            await AssertQuery<Customer, Order>((cs, os) =>
                from c in cs
                join o in os on c.CustomerID equals o.CustomerID into orders
                select new { cust = c, ords = orders.Count() });
        }

        [Fact]
        public virtual async Task GroupJoin_default_if_empty()
        {
            await AssertQuery<Customer, Order>((cs, os) =>
                from c in cs
                join o in os on c.CustomerID equals o.CustomerID into orders
                from o1 in orders.DefaultIfEmpty()
                select new { c, o1 });
        }

        [Fact]
        public virtual async Task SelectMany_customer_orders()
        {
            await AssertQuery<Customer, Order>((cs, os) =>
                from c in cs
                from o in os
                where c.CustomerID == o.CustomerID
                select new { c.ContactName, o.OrderID });
        }

        // TODO: Composite keys, slow..

        //        [Fact]
        //        public virtual async Task Multiple_joins_with_join_conditions_in_where()
        //        {
        //            await AssertQuery<Customer, Order, OrderDetail>((cs, os, ods) =>
        //                from c in cs
        //                from o in os.OrderBy(o1 => o1.OrderID).Take(10)
        //                from od in ods
        //                where o.CustomerID == c.CustomerID
        //                    && o.OrderID == od.OrderID
        //                where c.CustomerID == "ALFKI"
        //                select od.ProductID,
        //                assertOrder: true);
        //        }
        //        [Fact]
        //
        //        public virtual async Task TestMultipleJoinsWithMissingJoinCondition()
        //        {
        //            await AssertQuery<Customer, Order, OrderDetail>((cs, os, ods) =>
        //                from c in cs
        //                from o in os
        //                from od in ods
        //                where o.CustomerID == c.CustomerID
        //                where c.CustomerID == "ALFKI"
        //                select od.ProductID
        //                );
        //        }

        [Fact]
        public virtual async Task OrderBy()
        {
            await AssertQuery<Customer>(cs =>
                cs.OrderBy(c => c.CustomerID),
                assertOrder: true);
        }

        [Fact]
        public virtual async Task OrderBy_client_mixed()
        {
            await AssertQuery<Customer>(cs =>
                cs.OrderBy(c => c.IsLondon).ThenBy(c => c.CompanyName),
                assertOrder: true);
        }

        [Fact]
        public virtual async Task OrderBy_multiple_queries()
        {
            await AssertQuery<Customer, Order>((cs, os) =>
                from c in cs
                join o in os on new Foo { Bar = c.CustomerID } equals new Foo { Bar = o.CustomerID }
                orderby c.IsLondon, o.OrderDate
                select new { c, o });
        }

        [Fact]
        public virtual async Task OrderBy_shadow()
        {
            await AssertQuery<Employee>(es =>
                es.OrderBy(e => e.Property<string>("Title")).ThenBy(e => e.EmployeeID),
                assertOrder: true);
        }

        [Fact]
        public virtual async Task OrderBy_ThenBy_predicate()
        {
            await AssertQuery<Customer>(cs =>
                cs.Where(c => c.City == "London")
                    .OrderBy(c => c.City)
                    .ThenBy(c => c.CustomerID),
                assertOrder: true);
        }

        [Fact]
        public virtual async Task OrderBy_correlated_subquery_lol()
        {
            await AssertQuery<Customer>(cs =>
                from c in cs
                orderby cs.Any(c2 => c2.CustomerID == c.CustomerID)
                select c);
        }

        [Fact]
        public virtual async Task OrderBy_Select()
        {
            await AssertQuery<Customer>(cs =>
                cs.OrderBy(c => c.CustomerID)
                    .Select(c => c.ContactName),
                assertOrder: true);
        }

        [Fact]
        public virtual async Task OrderBy_multiple()
        {
            await AssertQuery<Customer>(cs =>
                cs.OrderBy(c => c.CustomerID)
                    // ReSharper disable once MultipleOrderBy
                    .OrderBy(c => c.Country)
                    .Select(c => c.City),
                assertOrder: true);
        }

        [Fact]
        public virtual async Task OrderBy_ThenBy()
        {
            await AssertQuery<Customer>(cs =>
                cs.OrderBy(c => c.CustomerID).ThenBy(c => c.Country).Select(c => c.City),
                assertOrder: true);
        }

        [Fact]
        public virtual async Task OrderByDescending()
        {
            await AssertQuery<Customer>(cs =>
                cs.OrderByDescending(c => c.CustomerID).Select(c => c.City),
                assertOrder: true);
        }

        [Fact]
        public virtual async Task OrderByDescending_ThenBy()
        {
            await AssertQuery<Customer>(cs =>
                cs.OrderByDescending(c => c.CustomerID).ThenBy(c => c.Country).Select(c => c.City),
                assertOrder: true);
        }

        [Fact]
        public virtual async Task OrderByDescending_ThenByDescending()
        {
            await AssertQuery<Customer>(cs =>
                cs.OrderByDescending(c => c.CustomerID).ThenByDescending(c => c.Country).Select(c => c.City),
                assertOrder: true);
        }

        [Fact]
        public virtual async Task OrderBy_Join()
        {
            await AssertQuery<Customer, Order>((cs, os) =>
                from c in cs.OrderBy(c => c.CustomerID)
                join o in os.OrderBy(o => o.OrderID) on c.CustomerID equals o.CustomerID
                select new { c.CustomerID, o.OrderID });
        }

        [Fact]
        public virtual async Task OrderBy_SelectMany()
        {
            await AssertQuery<Customer, Order>((cs, os) =>
                from c in cs.OrderBy(c => c.CustomerID)
                from o in os.OrderBy(o => o.OrderID)
                where c.CustomerID == o.CustomerID
                select new { c.ContactName, o.OrderID },
                assertOrder: true);
        }

        // TODO: Need to figure out how to do this 
        //        [Fact]
        //        public virtual async Task GroupBy_anonymous()
        //        {
        //            await AssertQuery<Customer>(cs =>
        //                cs.Select(c => new { c.City, c.CustomerID })
        //                    .GroupBy(a => a.City),
        //                assertOrder: true);
        //        }
        //
        //        [Fact]
        //        public virtual async Task GroupBy_anonymous_subquery()
        //        {
        //            await AssertQuery<Customer>(cs =>
        //                cs.Select(c => new { c.City, c.CustomerID })
        //                    .GroupBy(a => from c2 in cs select c2),
        //                assertOrder: true);
        //        }
        //
        //        [Fact]
        //        public virtual async Task GroupBy_nested_order_by_enumerable()
        //        {
        //            await AssertQuery<Customer>(cs =>
        //                cs.Select(c => new { c.City, c.CustomerID })
        //                    .OrderBy(a => a.City)
        //                    .GroupBy(a => a.City)
        //                    .Select(g => g.OrderBy(a => a.CustomerID)),
        //                assertOrder: true);
        //        }

        [Fact]
        public virtual async Task GroupBy_SelectMany()
        {
            await AssertQuery<Customer>(cs =>
                cs.GroupBy(c => c.City).SelectMany(g => g));
        }

        [Fact]
        public virtual async Task GroupBy_Sum()
        {
            await AssertQuery<Order>(os =>
                os.GroupBy(o => o.CustomerID).Select(g => g.Sum(o => o.OrderID)));
        }

        [Fact]
        public virtual async Task GroupBy_Count()
        {
            await AssertQuery<Order>(os =>
                os.GroupBy(o => o.CustomerID).Select(g => g.Count()));
        }

        [Fact]
        public virtual async Task GroupBy_LongCount()
        {
            await AssertQuery<Order>(os =>
                os.GroupBy(o => o.CustomerID).Select(g => g.LongCount()));
        }

        // TODO: #592
        //[Fact]
        public virtual async Task GroupBy_Sum_Min_Max_Avg()
        {
            await AssertQuery<Order>(os =>
                os.GroupBy(o => o.CustomerID).Select(g =>
                    new
                        {
                            Sum = g.Sum(o => o.OrderID),
                            Min = g.Min(o => o.OrderID),
                            Max = g.Max(o => o.OrderID),
                            Avg = g.Average(o => o.OrderID)
                        }));
        }

        // TODO: #592
        //[Fact]
        public virtual async Task GroupBy_with_result_selector()
        {
            await AssertQuery<Order>(os =>
                os.GroupBy(o => o.CustomerID, (k, g) =>
                    new
                        {
                            Sum = g.Sum(o => o.OrderID),
                            Min = g.Min(o => o.OrderID),
                            Max = g.Max(o => o.OrderID),
                            Avg = g.Average(o => o.OrderID)
                        }));
        }

        [Fact]
        public virtual async Task GroupBy_with_element_selector_sum()
        {
            await AssertQuery<Order>(os =>
                os.GroupBy(o => o.CustomerID, o => o.OrderID).Select(g => g.Sum()));
        }

        [Fact]
        public virtual async Task GroupBy_with_element_selector()
        {
            await AssertQuery<Order>(os =>
                os.GroupBy(o => o.CustomerID, o => o.OrderID)
                    .OrderBy(g => g.Key)
                    .Select(g => g.OrderBy(o => o)),
                asserter:
                    (l2oResults, efResults) =>
                        {
                            var l2oObjects
                                = l2oResults
                                    .SelectMany(q1 => ((IEnumerable<int>)q1));

                            var efObjects
                                = efResults
                                    .SelectMany(q1 => ((IEnumerable<int>)q1));

                            Assert.Equal(l2oObjects, efObjects);
                        });
        }

        [Fact]
        public virtual async Task GroupBy_with_element_selector_sum_max()
        {
            await AssertQuery<Order>(os =>
                os.GroupBy(o => o.CustomerID, o => o.OrderID)
                    .Select(g => new { Sum = g.Sum(), Max = g.Max() }));
        }

        [Fact]
        public virtual async Task GroupBy_with_anonymous_element()
        {
            await AssertQuery<Order>(os =>
                os.GroupBy(o => o.CustomerID, o => new { o.OrderID })
                    .Select(g => g.Sum(x => x.OrderID)));
        }

        [Fact]
        public virtual async Task GroupBy_with_two_part_key()
        {
            await AssertQuery<Order>(os =>
                os.GroupBy(o => new { o.CustomerID, o.OrderDate })
                    .Select(g => g.Sum(o => o.OrderID)));
        }

        [Fact]
        public virtual async Task OrderBy_GroupBy()
        {
            await AssertQuery<Order>(os =>
                os.OrderBy(o => o.OrderID)
                    .GroupBy(o => o.CustomerID)
                    .Select(g => g.Sum(o => o.OrderID)));
        }

        [Fact]
        public virtual async Task OrderBy_GroupBy_SelectMany()
        {
            await AssertQuery<Order>(os =>
                os.OrderBy(o => o.OrderID)
                    .GroupBy(o => o.CustomerID)
                    .SelectMany(g => g));
        }

        [Fact]
        public virtual async Task Sum_with_no_arg()
        {
            await AssertQuery<Order>(os => os.Select(o => o.OrderID).SumAsync());
        }

        [Fact]
        public virtual async Task Sum_with_arg()
        {
            await AssertQuery<Order>(os => os.SumAsync(o => o.OrderID));
        }

        [Fact]
        public virtual async Task Min_with_no_arg()
        {
            await AssertQuery<Order>(os => os.Select(o => o.OrderID).MinAsync());
        }

        [Fact]
        public virtual async Task Min_with_arg()
        {
            await AssertQuery<Order>(os => os.MinAsync(o => o.OrderID));
        }

        [Fact]
        public virtual async Task Max_with_no_arg()
        {
            await AssertQuery<Order>(os => os.Select(o => o.OrderID).MaxAsync());
        }

        [Fact]
        public virtual async Task Max_with_arg()
        {
            await AssertQuery<Order>(os => os.MaxAsync(o => o.OrderID));
        }

        [Fact]
        public virtual async Task Count_with_no_predicate()
        {
            await AssertQuery<Order>(os => os.CountAsync());
        }

        [Fact]
        public virtual async Task Count_with_predicate()
        {
            await AssertQuery<Order>(os =>
                os.CountAsync(o => o.CustomerID == "ALFKI"));
        }

        [Fact]
        public virtual async Task Distinct()
        {
            await AssertQuery<Customer>(cs => cs.Distinct());
        }

        [Fact]
        public virtual async Task Distinct_Scalar()
        {
            await AssertQuery<Customer>(cs =>
                cs.Select(c => c.City).Distinct());
        }

        [Fact]
        public virtual async Task OrderBy_Distinct()
        {
            await AssertQuery<Customer>(cs =>
                cs.OrderBy(c => c.CustomerID).Select(c => c.City).Distinct());
        }

        [Fact]
        public virtual async Task Distinct_OrderBy()
        {
            await AssertQuery<Customer>(cs =>
                cs.Select(c => c.City).Distinct().OrderBy(c => c),
                assertOrder: true);
        }

        [Fact]
        public virtual async Task Distinct_GroupBy()
        {
            await AssertQuery<Order>(os =>
                os.Distinct()
                    .GroupBy(o => o.CustomerID)
                    .OrderBy(g => g.Key)
                    .Select(g => new { g.Key, c = g.Count() }),
                assertOrder: true);
        }

        [Fact]
        public virtual async Task GroupBy_Distinct()
        {
            await AssertQuery<Order>(os =>
                os.GroupBy(o => o.CustomerID).Distinct().Select(g => g.Key));
        }

        [Fact]
        public virtual async Task Distinct_Count()
        {
            await AssertQuery<Customer>(cs => cs.Distinct().CountAsync());
        }

        [Fact]
        public virtual async Task Select_Distinct_Count()
        {
            await AssertQuery<Customer>(cs =>
                cs.Select(c => c.City).Distinct().CountAsync());
        }

        [Fact]
        public virtual async Task Select_Select_Distinct_Count()
        {
            await AssertQuery<Customer>(cs =>
                cs.Select(c => c.City).Select(c => c).Distinct().CountAsync());
        }

        [Fact]
        public virtual async Task Single_Throws()
        {
            await Assert.ThrowsAsync<InvalidOperationException>(() =>
                AssertQuery<Customer>(cs => cs.SingleAsync()));
        }

        [Fact]
        public virtual async Task Single_Predicate()
        {
            await AssertQuery<Customer>(
                cs => cs.SingleAsync(c => c.CustomerID == "ALFKI"));
        }

        public virtual async Task Single_Predicate_Cancellation(CancellationToken cancellationToken)
        {
            await AssertQuery<Customer>(
                cs => cs.SingleAsync(c => c.CustomerID == "ALFKI", cancellationToken));
        }

        [Fact]
        public virtual async Task Where_Single()
        {
            await AssertQuery<Customer>(
                // ReSharper disable once ReplaceWithSingleCallToSingle
                cs => cs.Where(c => c.CustomerID == "ALFKI").SingleAsync());
        }

        [Fact]
        public virtual async Task SingleOrDefault_Throws()
        {
            await Assert.ThrowsAsync<InvalidOperationException>(() =>
                AssertQuery<Customer>(cs => cs.SingleOrDefaultAsync()));
        }

        [Fact]
        public virtual async Task SingleOrDefault_Predicate()
        {
            await AssertQuery<Customer>(
                cs => cs.SingleOrDefaultAsync(c => c.CustomerID == "ALFKI"));
        }

        [Fact]
        public virtual async Task Where_SingleOrDefault()
        {
            await AssertQuery<Customer>(
                // ReSharper disable once ReplaceWithSingleCallToSingleOrDefault
                cs => cs.Where(c => c.CustomerID == "ALFKI").SingleOrDefaultAsync());
        }

        [Fact]
        public virtual async Task First()
        {
            await AssertQuery<Customer>(
                cs => cs.OrderBy(c => c.ContactName).FirstAsync());
        }

        [Fact]
        public virtual async Task First_Predicate()
        {
            await AssertQuery<Customer>(
                cs => cs.OrderBy(c => c.ContactName).FirstAsync(c => c.City == "London"));
        }

        [Fact]
        public virtual async Task Where_First()
        {
            await AssertQuery<Customer>(
                // ReSharper disable once ReplaceWithSingleCallToFirst
                cs => cs.OrderBy(c => c.ContactName).Where(c => c.City == "London").FirstAsync());
        }

        [Fact]
        public virtual async Task FirstOrDefault()
        {
            await AssertQuery<Customer>(
                cs => cs.OrderBy(c => c.ContactName).FirstOrDefaultAsync());
        }

        [Fact]
        public virtual async Task FirstOrDefault_Predicate()
        {
            await AssertQuery<Customer>(
                cs => cs.OrderBy(c => c.ContactName).FirstOrDefaultAsync(c => c.City == "London"));
        }

        [Fact]
        public virtual async Task Where_FirstOrDefault()
        {
            await AssertQuery<Customer>(
                // ReSharper disable once ReplaceWithSingleCallToFirstOrDefault
                cs => cs.OrderBy(c => c.ContactName).Where(c => c.City == "London").FirstOrDefaultAsync());
        }

        [Fact]
        public virtual async Task Last()
        {
            await AssertQuery<Customer>(
                cs => cs.OrderBy(c => c.ContactName).LastAsync());
        }

        [Fact]
        public virtual async Task Last_when_no_order_by()
        {
            await AssertQuery<Customer>(
                // ReSharper disable once ReplaceWithSingleCallToLast
                cs => cs.Where(c => c.CustomerID == "ALFKI").LastAsync());
        }

        [Fact]
        public virtual async Task Last_Predicate()
        {
            await AssertQuery<Customer>(
                cs => cs.OrderBy(c => c.ContactName).LastAsync(c => c.City == "London"));
        }

        [Fact]
        public virtual async Task Where_Last()
        {
            await AssertQuery<Customer>(
                // ReSharper disable once ReplaceWithSingleCallToLast
                cs => cs.OrderBy(c => c.ContactName).Where(c => c.City == "London").LastAsync());
        }

        [Fact]
        public virtual async Task LastOrDefault()
        {
            await AssertQuery<Customer>(
                cs => cs.OrderBy(c => c.ContactName).LastOrDefaultAsync());
        }

        [Fact]
        public virtual async Task LastOrDefault_Predicate()
        {
            await AssertQuery<Customer>(
                cs => cs.OrderBy(c => c.ContactName).LastOrDefaultAsync(c => c.City == "London"));
        }

        [Fact]
        public virtual async Task Where_LastOrDefault()
        {
            await AssertQuery<Customer>(
                // ReSharper disable once ReplaceWithSingleCallToLastOrDefault
                cs => cs.OrderBy(c => c.ContactName).Where(c => c.City == "London").LastOrDefaultAsync());
        }

        [Fact]
        public virtual async Task String_StartsWith_Literal()
        {
            await AssertQuery<Customer>(cs => cs.Where(c => c.ContactName.StartsWith("M")));
        }

        [Fact]
        public virtual async Task String_StartsWith_Identity()
        {
            await AssertQuery<Customer>(cs => cs.Where(c => c.ContactName.StartsWith(c.ContactName)));
        }

        [Fact]
        public virtual async Task String_StartsWith_Column()
        {
            await AssertQuery<Customer>(cs => cs.Where(c => c.ContactName.StartsWith(c.ContactName)));
        }

        [Fact]
        public virtual async Task String_StartsWith_MethodCall()
        {
            await AssertQuery<Customer>(cs => cs.Where(c => c.ContactName.StartsWith(LocalMethod1())));
        }

        [Fact]
        public virtual async Task String_EndsWith_Literal()
        {
            await AssertQuery<Customer>(cs => cs.Where(c => c.ContactName.EndsWith("b")));
        }

        [Fact]
        public virtual async Task String_EndsWith_Identity()
        {
            await AssertQuery<Customer>(cs => cs.Where(c => c.ContactName.EndsWith(c.ContactName)));
        }

        [Fact]
        public virtual async Task String_EndsWith_Column()
        {
            await AssertQuery<Customer>(cs => cs.Where(c => c.ContactName.EndsWith(c.ContactName)));
        }

        [Fact]
        public virtual async Task String_EndsWith_MethodCall()
        {
            await AssertQuery<Customer>(cs => cs.Where(c => c.ContactName.EndsWith(LocalMethod2())));
        }

        [Fact]
        public virtual async Task Contains_with_subquery()
        {
            await AssertQuery<Customer, Order>((cs, os) =>
                cs.Where(c => os.Select(o => o.CustomerID).Contains(c.CustomerID)));
        }

        [Fact]
        public virtual async Task Contains_with_local_collection()
        {
            string[] ids = new[] { "ABCDE", "ALFKI" };
            await AssertQuery<Customer>(cs =>
                cs.Where(c => ids.Contains(c.CustomerID)));
        }

        [Fact]
        public virtual async Task Contains_top_level()
        {
            await AssertQuery<Customer>(cs =>
                cs.Select(c => c.CustomerID).ContainsAsync("ALFKI"));
        }

        private static string LocalMethod1()
        {
            return "M";
        }

        private static string LocalMethod2()
        {
            return "m";
        }
        
        protected NorthwindContext CreateContext()
        {
            return Fixture.CreateContext();
        }

        protected AsyncQueryTestBase(TFixture fixture)
        {
            Fixture = fixture;
        }

        protected TFixture Fixture { get; private set; }

        private async Task<int> AssertQuery<TItem>(
            Func<IQueryable<TItem>, Task<int>> query,
            bool assertOrder = false)
            where TItem : class
        {
            using (var context = CreateContext())
            {
                return AssertResults(
                    new[] { await query(NorthwindData.Set<TItem>()) },
                    new[] { await query(context.Set<TItem>()) },
                    assertOrder);
            }
        }

        private async Task<int> AssertQuery<TItem>(
            Func<IQueryable<TItem>, Task<bool>> query,
            bool assertOrder = false)
            where TItem : class
        {
            using (var context = CreateContext())
            {
                return AssertResults(
                    new[] { await query(NorthwindData.Set<TItem>()) },
                    new[] { await query(context.Set<TItem>()) },
                    assertOrder);
            }
        }

        private async Task<int> AssertQuery<TItem>(
            Func<IQueryable<TItem>, Task<TItem>> query,
            bool assertOrder = false)
            where TItem : class
        {
            using (var context = CreateContext())
            {
                return AssertResults(
                    new[] { await query(NorthwindData.Set<TItem>()) },
                    new[] { await query(context.Set<TItem>()) },
                    assertOrder);
            }
        }

        private async Task<int> AssertQuery<TItem1, TItem2>(
            Func<IQueryable<TItem1>, IQueryable<TItem2>, Task<object>> query,
            bool assertOrder = false)
            where TItem1 : class
            where TItem2 : class
        {
            using (var context = CreateContext())
            {
                return AssertResults(
                    new[] { await query(NorthwindData.Set<TItem1>(), NorthwindData.Set<TItem2>()) },
                    new[] { await query(context.Set<TItem1>(), context.Set<TItem2>()) },
                    assertOrder);
            }
        }

        private async Task<int> AssertQuery<TItem>(
            Func<IQueryable<TItem>, IQueryable<IQueryable<object>>> query,
            bool assertOrder = false,
            Action<IList<IQueryable<object>>, IList<IQueryable<object>>> asserter = null)
            where TItem : class
        {
            using (var context = CreateContext())
            {
                return AssertResults(
                    query(NorthwindData.Set<TItem>()).ToArray(),
                    await query(context.Set<TItem>()).ToArrayAsync(),
                    assertOrder,
                    asserter);
            }
        }

        private async Task<int> AssertQuery<TItem>(
            Func<IQueryable<TItem>, IQueryable<object>> query,
            bool assertOrder = false,
            Action<IList<object>, IList<object>> asserter = null)
            where TItem : class
        {
            using (var context = CreateContext())
            {
                return AssertResults(
                    query(NorthwindData.Set<TItem>()).ToArray(),
                    await query(context.Set<TItem>()).ToArrayAsync(),
                    assertOrder,
                    asserter);
            }
        }

        private async Task<int> AssertQuery<TItem1, TItem2>(
            Func<IQueryable<TItem1>, IQueryable<TItem2>, IQueryable<object>> query,
            bool assertOrder = false,
            Action<IList<object>, IList<object>> asserter = null)
            where TItem1 : class
            where TItem2 : class
        {
            using (var context = CreateContext())
            {
                return AssertResults(
                    query(NorthwindData.Set<TItem1>(), NorthwindData.Set<TItem2>()).ToArray(),
                    await query(context.Set<TItem1>(), context.Set<TItem2>()).ToArrayAsync(),
                    assertOrder,
                    asserter);
            }
        }

        private async Task<int> AssertQuery<TItem1, TItem2, TItem3>(
            Func<IQueryable<TItem1>, IQueryable<TItem2>, IQueryable<TItem3>, IQueryable<object>> query,
            bool assertOrder = false)
            where TItem1 : class
            where TItem2 : class
            where TItem3 : class
        {
            using (var context = CreateContext())
            {
                return AssertResults(
                    query(NorthwindData.Set<TItem1>(), NorthwindData.Set<TItem2>(), NorthwindData.Set<TItem3>()).ToArray(),
                    await query(context.Set<TItem1>(), context.Set<TItem2>(), context.Set<TItem3>()).ToArrayAsync(),
                    assertOrder);
            }
        }

        private async Task<int> AssertQuery<TItem>(
            Func<IQueryable<TItem>, IQueryable<int>> query, bool assertOrder = false)
            where TItem : class
        {
            using (var context = CreateContext())
            {
                return AssertResults(
                    query(NorthwindData.Set<TItem>()).ToArray(),
                    await query(context.Set<TItem>()).ToArrayAsync(),
                    assertOrder);
            }
        }

        private async Task<int> AssertQuery<TItem>(
            Func<IQueryable<TItem>, IQueryable<long>> query, bool assertOrder = false)
            where TItem : class
        {
            using (var context = CreateContext())
            {
                return AssertResults(
                    query(NorthwindData.Set<TItem>()).ToArray(),
                    await query(context.Set<TItem>()).ToArrayAsync(),
                    assertOrder);
            }
        }

        private async Task<int> AssertQuery<TItem>(
            Func<IQueryable<TItem>, IQueryable<bool>> query, bool assertOrder = false)
            where TItem : class
        {
            using (var context = CreateContext())
            {
                return AssertResults(
                    query(NorthwindData.Set<TItem>()).ToArray(),
                    await query(context.Set<TItem>()).ToArrayAsync(),
                    assertOrder);
            }
        }

        private static int AssertResults<T>(
            IList<T> l2oItems,
            IList<T> efItems,
            bool assertOrder,
            Action<IList<T>, IList<T>> asserter = null)
        {
            Assert.Equal(l2oItems.Count, efItems.Count);

            if (asserter != null)
            {
                asserter(l2oItems, efItems);
            }
            else
            {
                if (assertOrder)
                {
                    Assert.Equal(l2oItems, efItems);
                }
                else
                {
                    foreach (var l2oItem in l2oItems)
                    {
                        if (!efItems.Contains(l2oItem))
                        {
                            Assert.True(
                                false,
                                string.Format(
                                    "\r\nL2o item: [{0}] not found in EF results:\r\n{1}",
                                    l2oItem,
                                    string.Join("\r\n", efItems)));
                        }
                    }
                }
            }

            return l2oItems.Count;
        }
    }
}
